<?php
/**
 * @copyright Copyright (c) 2019 Dominik Ach <da@infodatacom.de>
 *
 * @author Dominik Ach <da@infodatacom.de>
 * @author Arthur Schiwon <blizzz@arthur-schiwon.de>
 * @author Carl Schwan <carl@carlschwan.eu>
 * @author Maximilian Ruta <mr@xtain.net>
 * @author Jonathan Treffler <mail@jonathan-treffler.de>
 * @author Giuliano Mele <giuliano.mele@verdigado.com>
 *
 * @license GNU AGPL version 3 or any later version
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

namespace OCA\User_SAML;

use OC\BackgroundJob\JobList;
use OC\Hooks\PublicEmitter;
use OCA\User_SAML\Jobs\MigrateGroups;
use OCP\IConfig;
use OCP\IDBConnection;
use OCP\IGroup;
use OCP\IGroupManager;
use OCP\IUser;
use OCP\IUserManager;

class GroupManager {
	public const LOCAL_GROUPS_CHECK_FOR_MIGRATION = 'localGroupsCheckForMigration';

	/**
	 * @var IDBConnection $db
	 */
	protected $db;

	/**
	 * @var GroupDuplicateChecker
	 */
	protected $duplicateChecker;
	/** @var IGroupManager */
	private $groupManager;
	/** @var IUserManager */
	private $userManager;
	/** @var GroupBackend */
	private $ownGroupBackend;
	/** @var IConfig */
	private $config;
	/** @var JobList */
	private $jobList;
	/** @var SAMLSettings */
	private $settings;


	public function __construct(
		IDBConnection $db,
		GroupDuplicateChecker $duplicateChecker,
		IGroupManager $groupManager,
		IUserManager $userManager,
		GroupBackend $ownGroupBackend,
		IConfig $config,
		JobList $jobList,
		SAMLSettings $settings
	) {
		$this->db = $db;
		$this->duplicateChecker = $duplicateChecker;
		$this->groupManager = $groupManager;
		$this->userManager = $userManager;
		$this->ownGroupBackend = $ownGroupBackend;
		$this->config = $config;
		$this->jobList = $jobList;
		$this->settings = $settings;
	}

	private function getGroupsToRemove(array $samlGroups, array $assignedGroups): array {
		$groupsToRemove = [];
		foreach ($assignedGroups as $group) {
			// if group is not supplied by SAML and group has SAML backend
			if (!in_array($group->getGID(), $samlGroups) && $this->hasSamlBackend($group)) {
				$groupsToRemove[] = $group->getGID();
			}
		}
		return $groupsToRemove;
	}

	private function getGroupsToAdd(array $samlGroups, array $assignedGroupIds): array {
		$groupsToAdd = [];
		foreach ($samlGroups as $group) {
			// if user is not assigend to the group or the provided group has a non SAML backend
			if (!in_array($group, $assignedGroupIds) || !$this->hasSamlBackend($this->groupManager->get($group))) {
				$groupsToAdd[] = $group;
			}
		}
		return $groupsToAdd;
	}

	public function replaceGroups(string $uid, array $samlGroups): void {
		$user = $this->userManager->get($uid);
		if ($user === null) {
			return;
		}
		$this->translateGroupToIds($samlGroups);
		$assignedGroups = $this->groupManager->getUserGroups($user);
		$assignedGroupIds = array_map(function (IGroup $group) {
			return $group->getGID();
		}, $assignedGroups);
		$groupsToRemove = $this->getGroupsToRemove($samlGroups, $assignedGroups);
		$groupsToAdd = $this->getGroupsToAdd($samlGroups, $assignedGroupIds);
		$this->removeGroups($user, $groupsToRemove);
		$this->addGroups($user, $groupsToAdd);
	}

	protected function translateGroupToIds(array &$samlGroups): void {
		array_walk($samlGroups, function (&$gid) {
			$altGid = $this->ownGroupBackend->groupExistsWithDifferentGid($gid);
			if ($altGid !== null) {
				$gid = $altGid;
			}
		});
	}

	public function removeGroups(IUser $user, array $groupIds): void {
		foreach ($groupIds as $gid) {
			$this->removeGroup($user, $gid);
		}
	}

	public function removeGroup(IUser $user, string $gid): void {
		$group = $this->groupManager->get($gid);
		if ($group === null) {
			return;
		}
		$this->ownGroupBackend->removeFromGroup($user->getUID(), $group->getGID());
		if ($this->ownGroupBackend->countUsersInGroup($gid) === 0) {
			$this->ownGroupBackend->deleteGroup($group->getGID());
		}
	}

	public function addGroups(IUser $user, $groupIds): void {
		foreach ($groupIds as $gid) {
			$this->addGroup($user, $gid);
		}
	}

	public function addGroup(IUser $user, string $gid): void {
		try {
			$group = $this->findGroup($gid);
		} catch (\RuntimeException $e) {
			if ($e->getCode() === 1) {
				$group = $this->createGroupInBackend($gid);
			} elseif ($e->getCode() === 2) {
				//FIXME: probably need config flag. Previous to 17, gid was used as displayname
				$providerId = $this->settings->getProviderId();
				$settings = $this->settings->get($providerId);
				$groupPrefix = $settings['saml-attribute-mapping-group_mapping_prefix'] ?? 'SAML_';
				$group = $this->createGroupInBackend($groupPrefix . $gid, $gid);
			} else {
				throw $e;
			}
		}

		$group->addUser($user);
	}

	protected function createGroupInBackend(string $gid, ?string $originalGid = null): ?IGroup {
		if ($this->groupManager instanceof PublicEmitter) {
			$this->groupManager->emit('\OC\Group', 'preCreate', array($gid));
		}
		if (!$this->ownGroupBackend->createGroup($gid, $originalGid ?? $gid)) {
			return null;
		}

		$group = $this->groupManager->get($gid);
		if ($this->groupManager instanceof PublicEmitter) {
			$this->groupManager->emit('\OC\Group', 'postCreate', array($group));
		}

		return $group;
	}

	protected function findGroup(string $gid): IGroup {
		$migrationWhiteList = $this->config->getAppValue(
			'user_saml',
			GroupManager::LOCAL_GROUPS_CHECK_FOR_MIGRATION,
			null
		);
		$strictBackendCheck = null === $migrationWhiteList;
		if ($migrationWhiteList !== null) {
			$migrationWhiteList = \json_decode($migrationWhiteList, true);
		}
		if (!$strictBackendCheck && in_array($gid, $migrationWhiteList['groups'], true)) {
			$group = $this->groupManager->get($gid);
			if ($group === null) {
				//FIXME: specific Exception and/or constant error code
				throw new \RuntimeException('Group not found', 1);
			}
			return $group;
		}
		$group = $this->groupManager->get($gid);
		if ($group === null) {
			//FIXME: specific Exception and/or constant error code
			throw new \RuntimeException('Group not found', 1);
		}
		if ($this->hasSamlBackend($group)) {
			return $group;
		}

		$altGid = $this->ownGroupBackend->groupExistsWithDifferentGid($gid);
		if ($altGid) {
			return $this->groupManager->get($altGid);
		}

		//FIXME: specific Exception and/or constant error code
		throw new \RuntimeException('Non-migratable duplicate found', 2);
	}

	protected function hasSamlBackend(IGroup $group): bool {
		return in_array('user_saml', $group->getBackendNames());
	}

	public function evaluateGroupMigrations(array $groups): void {
		$candidateInfo = $this->config->getAppValue('user_saml', self::LOCAL_GROUPS_CHECK_FOR_MIGRATION, null);
		if ($candidateInfo === null) {
			return;
		}
		$candidateInfo = \json_decode($candidateInfo, true);
		if (!isset($candidateInfo['dropAfter']) || $candidateInfo['dropAfter'] < time()) {
			$this->config->deleteAppValue('user_saml', self::LOCAL_GROUPS_CHECK_FOR_MIGRATION);
			return;
		}

		$this->jobList->add(MigrateGroups::class, ['gids' => $groups]);
	}
}
